/*******************************************
Program                 : AC-Cal Lite
Compiler                : WinAVR
System Clock            : 16 MHz
*******************************************/
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
//******************************************
//    
long int BUFFER;
#define DDS_CLK 25006000;
const long int DELTA[7] = {1000000,100000,10000,1000,100,10,1};
int MODE, CUR_POS;
#define M_1V        0
#define M_10V       1
#define M_Cal       2
//******************************************
//  ,   EEPROM

//******************************************
//   

//******************************************
//    
#define k_Up        2
#define k_Down      1
#define k_Left      3
#define k_Right     4
#define k_Center    0
//******************************************
#define _KeyBufferMax 10                //  FIFO  
#define _MaxKeys 5                      //     
volatile char _KeyBuffer[_KeyBufferMax];//    
volatile int  _KeyBufferPos;            //     
const unsigned char _KeyLoADC[5] = {29,76,111,146,184}; 
const unsigned char _KeyHiADC[5] = {76,111,146,184,230};
//******************************************
volatile char isKeyPressed[_MaxKeys];   //    
//******************************************
//  
#define Beep_On()  TCCR1A|=(1<<COM1A0); //    OC1A
#define Beep_Off() TCCR1A =TCCR1A&~(1<<COM1A0);
//******************************************
//  
#define Select_1V()     PORTB |= 0b00000001
#define Select_10V()    PORTB &= 0b11111110
#define Selfcal_On()    PORTB |= 0b00000100
#define Selfcal_Off()   PORTB &= 0b11111011
//******************************************
//     DDS AD9833
#define DDS_SCLK_SET()  PORTD |= 0b01000000
#define DDS_SCLK_CLR()  PORTD &= 0b10111111
#define DDS_FSYNC_SET() PORTD |= 0b10000000 
#define DDS_FSYNC_CLR() PORTD &= 0b01111111 
#define DDS_SDATA(x)    PORTD = (PORTD & 0b11011111)|((x & 0b00000001)<<5)
//******************************************
//      AD7794
#define ADC_SCLK_SET()  PORTD |= 0b00010000
#define ADC_SCLK_CLR()  PORTD &= 0b11101111
#define ADC_DOUT()      (PORTD & 0b11111011)>>2
#define ADC_DIN(x)      PORTD = (PORTD & 0b11110111)|((x & 0b00000001)<<3)
//******************************************
//    AD7794
#define ADC_COMMREG     0b00000000
#define ADC_STATREG     0b00000000
#define ADC_MODEREG     0b00001000
#define ADC_CONFREG     0b00010000
#define ADC_DATAREG     0b00011000
//******************************************
//      LCD 
unsigned char BufferLCD[10];            
//******************************************
//    LCD MT-10T7
#define LCD_A0_SET()  PORTB |= 0b00010000
#define LCD_A0_CLR()  PORTB &= 0b11101111
#define LCD_WR1_SET() PORTC |= 0b00010000
#define LCD_WR1_CLR() PORTC &= 0b11101111
#define LCD_DB_CLR()  PORTC &= 0b11110000
#define LCD_DB_SET(x) PORTC |= (x)
/*******************************************
//   LCD MT-10T7
   a
  ---        0b00000001  g
 |   |       0b00000010  e
f|   |b      0b00000100  d
 | g |       0b00001000  a
  ---        0b00010000  h
 |   |       0b00100000  b
e|   |c      0b01000000  c
 | d |       0b10000000  f
  --- .h
*******************************************/
//    
const unsigned char Dig[16] = {          //   
0b11101110, // 0
0b01100000, // 1
0b00101111, // 2
0b01101101, // 3
0b11100001, // 4
0b11001101, // 5
0b11001111, // 6
0b11101000, // 7
0b11101111, // 8
0b11101101, // 9
0b11101011, // A
0b11000111, // b
0b10001110, // C
0b01100111, // d
0b10001111, // E
0b10001011};// F
#define s_point  0b00010000                 //    
#define s_minus  0b00000001    
#define s_space  0b00000000    
#define s_A      0b11101011    
#define s_b      0b11000111    
#define s_C      0b10001110    
#define s_d      0b01100111    
#define s_E      0b10001111    
#define s_F      0b10001011    
#define s__      0b00000100    
#define s_tilde  0b00001000    
#define s_degree 0b10101001    
#define s_c      0b00000111    
#define s_G      0b11001110    
#define s_H      0b11100011
#define s_I      0b01100000
#define s_L      0b10000110
#define s_l      0b00000110
#define s_n      0b01000011
#define s_o      0b01000111
#define s_O      0b11101110
#define s_P      0b10101011
#define s_R      0b10001010
#define s_S      0b11001101
#define s_r      0b00000011
#define s_t      0b10000111
#define s_U      0b11100110
#define s_u      0b01000110
#define s_Y      0b11100101
//******************************************
//      
//******************************************
void LCD_Addr (unsigned char a)
{
    LCD_A0_CLR();           // A0=0
    LCD_DB_CLR();
    LCD_DB_SET(a);          //  4 
    LCD_WR1_SET();
    _delay_us(1);           // 
    LCD_WR1_CLR();
    _delay_us(2);           // 
}
//******************************************
//     
//******************************************
void LCD_Write (unsigned char b)
{
    LCD_A0_SET();           // A0=1
    LCD_DB_CLR();
    LCD_DB_SET(b & 0x0F);   //  4 
    LCD_WR1_SET();
    _delay_us(1);           // 
    LCD_WR1_CLR();
    _delay_us(2);           // 
    LCD_DB_CLR();
    LCD_DB_SET(b >> 4);     //  4 
    LCD_WR1_SET();
    _delay_us(1);           // 
    LCD_WR1_CLR();
    _delay_us(2);           // 
}
//******************************************
//     
//******************************************
void LCD_Clear (void)
{
    LCD_Addr(0x00);      //     
    for(int j=0; j<10; j++) LCD_Write(BufferLCD[j] = s_space);
}
//******************************************
//       
//******************************************
void LCD_Update (void)
{
    LCD_Addr(0x00);      //     
    for(int j=0; j<10; j++) LCD_Write(BufferLCD[j]); 
}
//****************************************** 
//    AD9833
//****************************************** 
void DDS_Write (int data)
{
    DDS_SCLK_SET();
    DDS_FSYNC_CLR();
    int tmp = data;
    for (char i = 1; i<=16; i++) {
        DDS_SDATA(tmp>>15);
        tmp = tmp << 1;
        _delay_us(1);
        DDS_SCLK_CLR();
        _delay_us(1);
        DDS_SCLK_SET();
        _delay_us(1);
    };    
    DDS_FSYNC_SET();
}
//******************************************
//    
//****************************************** 
void KeyClear (void)
{
    for (int i=0; i<_KeyBufferMax; i++) isKeyPressed[i] = 0; //    
    for (int i=0; i<_MaxKeys; i++) _KeyBuffer[i] = 0;        //   
    _KeyBufferPos = 0;
}
//******************************************
//       ?
//****************************************** 
unsigned char KeyPressed (void)
{
    return(_KeyBufferPos!=0);
}
//******************************************
//    
//****************************************** 
unsigned char GetKey (void)
{
    if (_KeyBufferPos==0) return(0);    //   ,   0
    else {
      char _key = _KeyBuffer[0];        //          1 
      for (int i=0; i<_KeyBufferMax-1; i++) _KeyBuffer[i] = _KeyBuffer[i+1];
      _KeyBufferPos--;
      return(_key);
    }
}
//******************************************
//       
//****************************************** 
unsigned char ReadKey (void)
{
    while (!KeyPressed());    
    return(GetKey());
}
//******************************************
//      
//****************************************** 
ISR (TIMER2_COMP_vect)
{
    ADCSR |= (1<<ADSC);                             //  
    while (ADCSR & (1<<ADSC)) {};                   //    
    unsigned char key_raw = ADCH;     
    for (int idx=0; idx<5; idx++) {                 //     
      char key_state = (key_raw>_KeyLoADC[idx]) && (key_raw<_KeyHiADC[idx]);//    
      if (!isKeyPressed[idx] && (key_state)) {
        if (_KeyBufferPos < _KeyBufferMax) {        //    ,
          _KeyBuffer[_KeyBufferPos] = idx;          //      
          _KeyBufferPos++;
        };
      };
      isKeyPressed[idx]=key_state;                  //      
    };   
}
//******************************************
//     RS232 
//****************************************** 
void UART_Send (char c) 
{
    while (!(UCSRA & _BV(UDRE)));       //       
    UDR = c;                            //     UART
}                               
//******************************************
//    PHASE0 DDS
//******************************************
void SetPhase (int phase)
{
    DDS_Write(0b1100000000000000 | (phase & 0b0000111111111111));       
}
//******************************************
//    FREQ0 DDS
//******************************************
void SetFreq (long int freq)
{
    int64_t tmp = freq;
    tmp = (tmp<<28) / DDS_CLK;
    DDS_Write(0b0100000000000000 | (tmp & 0b0011111111111111));        // LSB
    DDS_Write(0b0100000000000000 | ((tmp>>14) & 0b0011111111111111));  // MSB
}
//******************************************
//  DDS   SLEEP1+RESET
//******************************************
void SetIdleMode (void)
{
    DDS_Write(0b0010000010000000);      // SLEEP1 = 1
    DDS_Write(0b0010000100000000);      // RESET = 1
}
//******************************************
//  DDS    
//******************************************
void SetNormalMode (void)
{
    DDS_Write(0b0010000000000000);
}
//******************************************
//   LCD   
//******************************************
void Display (void)
{
    long int tmp = BUFFER;
    for(unsigned char j=0; j<7; j++) {
      BufferLCD[6-j] = Dig[tmp % 10];
      tmp /= 10;}
    BufferLCD[CUR_POS] |= s_point;
    if (CUR_POS>0) BufferLCD[CUR_POS-1] |= s_point;
    switch (MODE) {
      case M_1V: 
        BufferLCD[8] = 0b01010000; BufferLCD[9] = 0b01000111;
      break;
      case M_10V: 
        BufferLCD[8] = 0b01000000; BufferLCD[9] = 0b01000111;
      break;
    };
    LCD_Update();
}
//******************************************
//  
//****************************************** 
void Init (void)
{
//  , , 
    DDRB  = 0b00010111;
    PORTB = 0b11101000;
    DDRD  = 0b11111010;
    PORTD = 0b11010101;
    DDRC  = 0b11011111;
    PORTC = 0b00000000;
    OCR1AH=0; OCR1AL=0x56;                  // 2000     
    TCCR1A =TCCR1A&~(1<<COM1A0);            //    OC1A
    TCCR1B|=(1<<WGM12)|(1<<CS11)|(1<<CS10); //  CTC.     64
    TCCR1A|=(1<<FOC1A);
    ADMUX = (0<<REFS1)|(1<<REFS0)|(1<<ADLAR)|(0<<MUX3)|(1<<MUX2)|(0<<MUX1)|(1<<MUX0);
    ADCSR = (1<<ADEN)|(0<<ADSC)|(0<<ADFR)|(0<<ADIE)|(1<<ADPS2)|(1<<ADPS1)|(1<<ADPS0);
//  RS232 
    UCSRB |=  _BV(TXEN) | _BV(RXEN);        //  Transmit  Receive
    UCSRC |= _BV(UCSZ1) | _BV(UCSZ0);       // 8-bit 
    UBRRH = 0; 
    UBRRL = 0x67;                           // 9600 
//    UCSRB |=  _BV(RXCIE);                 //    
    SREG |= _BV(SREG_I);                    //   
//  LCD
    LCD_Addr(0x0F);      //     
    LCD_Write(0xFF);    //   
    LCD_Clear();
//  
    ASSR = 0x00;                            //   2:   
    TCCR2 |= (1<<CS20)|(1<<CS21)|(1<<CS22); //   1024
    TCNT2 = 0x00;                           //   0
    OCR2 = 0xFE;                            //    
    TIMSK = (1<<OCIE2);                     //  
    KeyClear();
//  DDS
    SetNormalMode();
    SetFreq(0);
    SetPhase(0);
//    
    Select_1V();
    Selfcal_Off();
//   AD7794 
/*SendByteAD (0xff); //   (4  0xff)
SendByteAD (0xff);
SendByteAD (0xff);
SendByteAD (0xff);
-  1 ms   ;
SendByteAD (0x10); //      CONFIG 
SendByteAD (0x10); //   CONFIG  HB ( ,  1)
SendByteAD (0x10); //   CONFIG  LB (1  c )
SendByteAD (0x08); //      MODE 
SendByteAD (0x20); //   MODE  HB (Single Conversion)
SendByteAD (0x01); //   MODE  LB (Internal 64 kHz Clock, 4 ms conversion)
-  5 s       ;
SendByteAD (0x58); //     
ReadByteAD (); //    1 
ReadByteAD (); //    2 
ReadByteAD (); //    3  */
}
//******************************************
//    EEPROM
//******************************************
void Config_Load (void)
{
//    eeprom_read_block(&CAL_ZERO,&CFG_CAL_ZERO,32);
//    eeprom_read_block(&CAL_FS,&CFG_CAL_FS,32);
}
//******************************************
//    EEPROM
//******************************************
void Config_Update (void)
{
//    eeprom_update_block(&CAL_ZERO,&CFG_CAL_ZERO,32);
//    eeprom_update_block(&CAL_FS,&CFG_CAL_FS,32);
}
//******************************************
//     / / 
//******************************************
void Selector (void)
{
    Display();
    SetFreq(BUFFER);
    char SelIsDone = 0;
    while (!SelIsDone) {
      char k = ReadKey();    
      switch (k) {
        case k_Center:
          SelIsDone = 1;  
        break;
        case k_Left:
          if (CUR_POS>0) CUR_POS--;
        break;
        case k_Right:
          if (CUR_POS<6) CUR_POS++;
        break;
        case k_Up:
          if (BUFFER+DELTA[CUR_POS]<=1000000) BUFFER+=DELTA[CUR_POS];
        break;
        case k_Down:
          if (BUFFER-DELTA[CUR_POS]>=0) BUFFER-=DELTA[CUR_POS];
        break;
      }
      Display();
      SetFreq(BUFFER);
    }
}
//******************************************
int main (void)
{
    Config_Load();                  //    EEPROM
    cli();                          
    Init();                         //  
    sei();                          //  
    
    Beep_On(); _delay_ms(50); Beep_Off();
    MODE = M_1V;
    Select_1V();
    BUFFER=1000; CUR_POS = 6; 
    while(1) {                      //  
      Selector();
      LCD_Clear();
      BufferLCD[0] = s_C; BufferLCD[1] = s_L;        BufferLCD[3] = 0b01100000; 
      BufferLCD[4] = s_O; BufferLCD[6] = 0b01100001; BufferLCD[9] = 0b10000011;
      if (MODE==M_1V) BufferLCD[3] |= s_point;
      LCD_Update();
      char k = ReadKey();
      Beep_On(); _delay_ms(2); Beep_Off();
      if (k==k_Up) {
        switch (MODE) {
          case M_1V: 
            MODE = M_10V; Select_10V();
          break;
          case M_10V: 
            MODE = M_1V; Select_1V();
          break;
        };
        Beep_On(); _delay_ms(50); Beep_Off();
      };
      if (k==k_Down) BUFFER = 0; 
      if (k==k_Left)  {};
      if (k==k_Right) {}; 
      if (k==k_Center) {}
    };
    return 0;
}
